#!/usr/bin/env perl
#
# NAME:
#	dnsdeps.pl - DNS generate dependencies
#
# SYNOPSIS:
#	dnsdeps.pl [options] "file" ...
#
# DESCRIPTION:
#	This script recursively searches "file"s for $INCLUDE
#	directives and the zone defintions they contain. It handles
#	BIND-8 as well as BIND-4 config syntax.
#
#	Options:
#
#	-f "makefile"
#		Append dependencies to "makefile".
#		
#	-N	Just leave dependencies in '.depend' rather than
#		appending them to "makefile".
#
#	
# AUTHOR:
#	Simon J. Gerraty <sjg@crufty.net>
#

# RCSid:
#	$Id: dnsdeps.pl,v 1.9 2006/03/20 22:22:16 sjg Exp $
#
#	@(#)Copyright (c) 1993-2001 Simon J. Gerraty.
#	All rights reserved.
#	
#	This software is the property of Simon J. Gerraty.
#	It may be used only in accordance with the conditions outlined
#	in the accompanying LICENSE file. If your distribution does
#	not contain a LICENSE file, then you have no rights to use
#	this software and should contact sales@crufty.net to obtain a
#	licensed distribution.
#	
#	This software is provided "AS IS" in the hope that it will be
#	of use.  Since it is not possible to anticipate all possible
#	use of this software, there is absolutely NO WARRANTY.
#
#	Please forward bug reports and fixes to help@crufty.net
#

if ($0 =~ m,^(.*)/([^/]+)$,) {
  $Mydir = $1;
  $Myname = $2;
} else {
  $Mydir = '.';
  $Myname = $0;
}

sub source {
  local($file,@dirs) = @_;
  local($d);
  
  @dirs = ('.', $Mydir, '/etc') if (scalar(@dirs) == 0);

  foreach $d (@dirs) {
    if (-s "$d/$file") {
      do "$d/$file";
      last;
    }
  }
}

foreach $f ($MAKEFILE, 'makefile','Makefile') {
  if ($f ne '' && -s $f) {
    $opt_f = $f;
    last;
  }
}

&source('.dnsdepsrc.pl');

require 'getopts.pl';

&Getopts("anNV:s:w:o:I:D:b:f:");

$nest=0;

%Visited = ();

&main;
exit 0;

sub get_incs {
  local($file) = @_;

  local(@incs) = ();
  
  if (-e $file ) {
    if (open(I, "< $file")) {
      
      while (<I>) {
	next if (m,^\s*(;|//),); # skip comments
	if (m/include\s+"(\S+)"/ ||
	    m/\$INCLUDE\s+(\S+)/i) {
	  if ($1 eq $file) {
	    print STDERR "$Myname: bad format: $file includes $file\n";
	  } else {
	    push(@incs, $1);
	  }
	}
      }
      close I;
    }
  } else {
    # if it is a .soa file that is missing,
    # we know what it will contain
    # so we can pretend that it exists
    if ("./$file" =~ m,/([^/]+)\.soa,) {
      $file =~ s/\.soa/.db/;
      push(@incs, 'ns.list') if (-s 'ns.list');
      push(@incs, $file);
    } else {
      print STDERR "$Myname: cannot open: $file\n";
    }
  }
  @incs;
}

sub gen_dep {
  local(@files) = @_;
  local(@incs) = ();
  
  foreach $f (@files) {
    push(@incs, $f);
    push(@incs, &gen_dep(&get_incs($f)));
  }
  @incs;
}

sub gen_deps {
  local($file) = @_;
  local(@incs) = &get_incs($file);
  local($f,$lf);
  
  if (scalar(@incs) > 0) {
    print "\n$file:";
    $lf = '';
    foreach $f (sort(&gen_dep(@incs))) {
      next if ($f eq $lf);
      print " \\\n\t$f";
      $lf = $f;
    }
    print "\n\n";
  }
}

sub get_zones {
  local(*H) = @_;
  local($zfile,$ztype);
  local(@zones) = ();

  
  while (<H>) {
    next if (m,^\s*(;|//),); # skip comments
    if (m/zone\s+/) {
      $ztype = $zfile = '';
    } elsif (m/^primary\s+.*\s(\S+)$/) {
      push(@zones, $1);
    } elsif (m/file\s+"(\S+)"/) {
      $zfile = $1;
    } elsif (m/type\s+(\S+)/) {
      $ztype = $1;
    } elsif (m/include\s+"(\S+)"/) {
      $ifile = $1;
      if ($Visited{$ifile} eq '') {
	$Visited{$ifile} = 1;
	$nest++;
	push(@zones, $ifile);
	eval "if (open(Z$nest, '< $ifile')) {
push(\@zones, \&get_zones(*Z$nest));
	  close Z$nest; }";
	$nest--;
      } else {
	print STDERR "$Myname: skipping circular reference to: $ifile\n";
      }
    }
    if ($ztype =~ m/master/ && $zfile ne '') {
      push(@zones, $zfile);
      $ztype = $zfile = '';
    }
  }
  @zones;
}

sub main {
  @Zones = ();

  open(STDOUT, "> .depend") || die;
  
  foreach $file (@ARGV) {
    if (open(I, "< $file")) {
      @zones = &get_zones(*I);
      if (scalar(@zones) == 0) {
	&gen_deps($file);
      } else {
	foreach $z (@zones) {
	  &gen_deps($z);
	}
	push(@Zones, @zones);
      }
      close I;
    }
  }
  if (scalar(@Zones) > 0) {
    print "\n.zones:";
    foreach $z (@Zones) {
      print " \\\n\t$z";
    }
    print "\n\n";
  }
  close STDOUT;
  exit 0 if ($opt_N ne '');

  if ($opt_f eq '') {
    print STDERR "$Myname: no MAKEFILE specified.  Cannot add .depend\n";
    exit 1;
  }
  if (! -s $opt_f) {
    print STDERR "$Myname: $opt_f: missing or empty\n";
    exit 1;
  }
  if (-s '.depend') {
    #
    # Save the manually updated section of the makefile
    #
    # if make doesn't support include, then append our deps...
    chop($depended=`grep 'include.*\.depend' $opt_f`);
    exit 0 if ($depended ne '');
    if ($INCLUDE ne '') {
      open(STDOUT, ">> $opt_f") || die;
      if ($INCLUDE eq '.include') {
	print ".include \".depend\"\n";
      } elsif ($INCLUDE eq 'include') {
	print "include .depend\n";
      }
      close STDOUT;
      exit 0;
    }
    $TF="$opt_f.$$";
    open(K, "> $TF.k") || die;
    print K "
# DO NOT DELETE THIS LINE -- make depend depends on it
# Do not edit anything below, it was added automagically by $Myname.

";
    close K;
    
    system("sed '/^# DO NOT DELETE.*depend.*\$/,\$d' < $opt_f > $TF && mv $TF $opt_f && cat $TF.k .depend >> $opt_f; rm -f $TF $TF.k") || exit 0;
    exit 1;
  }
}

