#!/usr/bin/env perl
#
# NAME:
#	updrev.pl - update rev maps
#
# SYNOPSIS:
#	updrev.pl [options] ["named_dump"]
#
# DESCRIPTION:
#	This is a greatly simplified mkrev tool.  It uses input from
#	"named_dump" via getdata.pl(1).  We look at 'SOA','PTR' and 'A'
#	records.
#
#	'SOA' records tell us whether "ns" is authoritative for the
#	zone.
#
#	'PTR' records tell us whether a mapping already exists in which
#	case we leave it alone - provided that an A record for the
#	same fqdn also exists. This allows an initial set of maps to
#	be generated by this tool, then hand edited to reflect
#	preferences which will then be left alone. 
#
#	'A' records are used to generate new 'PTR' records on a first come
#	basis. We first determine the IN-ADDR zone that the address
#	belongs to and if "ns" is authoritative for that zone, or the
#	domain that the host is in, we generate a 'PTR' record.
#
#	When checking for authority, we work our way up through the
#	DNS hierarchy until we find an SOA record, or we hit "TOP_ZONE" or
#	"TOP_ARPA_ZONE", see comments below under options.
#	
#	Subsequent 'A' records for the same address, generate 'PTR'
#	records which are commented out.  This makes it easier for a
#	human hostmaster to pick a preferred PTR record from the
#	available choices.
#
#	If a file with the same name as a rev zone exists with a
#	'.inc' extention exists (eg. rev/203.12.250.inc) it will be
#	$INCLUEDed into the '.rev' zone.  This allows for non-PTR
#	records for delegations and such to be kept apart from the
#	automagically maintained '.rev' and '.soa' files.
#
#	For each new IN-ADDR zone we add an appropriate entry to
#	"ndboot" (use -N /dev/null to stop this).
#
#	The new reverse maps are only installed if they differ from
#	the previous version.  For this, we rely on the SOA records
#	being kept in a separate file to the zone data.	 The SOA
#	records are maintained by make(1) and updsoa.pl(8).
#
#	Options:
#
#	-d "dbg"
#		Turns on debugging output, "dbg" is a coma separated
#		list of assignments for example: sh,gen,ptr=5,soa=9
#		will produce lots of output.  Once must read the
#		source to really work it all out though.
#		
#	-c "config"
#		Read settings from "config".  This happens after
#		reading .dnsrc.pl and .updrev.pl, so overrides them
#		and the command line.
#
#	-n "ns"
#		"ns" is a coma separated list of nameservers for which
#		we will generate IN-ADDR zones.
#
#	-N "ndboot"
#		Add entries for new zones to "ndboot" (default is named.boot).
#		For bind-8 installations we use primary.zones which is
#		then included by named.conf.
#
#	-a "added"
#		List added .rev files in "added".  This is handy for
#		ensuring they are added to CVS.
#		
#	-D "data"
#		'getdata.pl' has already been run on 'named_dump.db'
#		and result is in "data".
#		
#	-A "TOP_ARPA_ZONE"
#		For sites that delegate in-addr zones, this is the
#		top level. We stop searching for an SOA record in parents
#		when we reach it. Can be set via "TOP_ARPA_ZONE" in
#		the environment.  The default is the exact
#		in-addr.arpa zone we are looking for - that is no
#		parents are checked.
#
#	-T "TOP_ZONE"
#		When searching for an SOA record for a domain or its
#		parent, stop when we reach "TOP_ZONE".	Default is the
#		parent domain of the 1st "ns". Can be set via
#		"TOP_ZONE" in the environment. 
#
#	-R "revdir"
#		Put the reverse maps in "revdir".  The files are named
#		for the network with a '.rev' extention and contain
#		only the PTR records.  A separate '.soa' file will be
#		maintained by make(1).
#
#	skip_fqdn:
#
#	If $A_hook is set to 'skip_fqdn', then all fqdns associated
#	with A records are checked against the regexps listed in
#	@Skip_fqdn_re.  This makes it simple to force certain names to
#	not be used as a PTR record.  For example:
#.nf
#
#	A host-le0.domain 192.168.200.1 
#	A host-le1.domain 192.168.201.1
#	A host.domain 192.168.200.1 
#	A host.domain 192.168.201.1
#.fi
#
#	Without any intervention we would get PTR records for the -le
#	names. Adding the following to '.updrevrc.pl' will cause the
#	-le names to be skipped.
#.nf
#
#	$A_hook = 'skip_fqdn';
#	push(@Skip_fqdn_re, '-(le|hme)\d+\.domain');
#.fi
#
#	CIDR:
#
#	By default 'updrev.pl' generates a zone file for each A, B
#	and C class network, regardless of the sub-netting in use.
#	This greatly reduces the number of zone files to be managed
#	and improves the performance of the tool.  To CIDR the
#	associative array %Mask needs to be populated with the
#	appropriate netmasks.  We check the netmask for each address
#	by looking up the A, B or C class network in %Mask, and
#	determine the prefix of the CIDR block that contains the
#	address.  Where the mask is a multiple of 8 bits things remain
#	simple.  In fact, by default the %Mask associative array
#	contains only entries for 'A', 'B' and 'C' which are looked up
#	if the specific network is not found.  Thus a site which wants
#	to subnet all networks on a 24 bit mask can simply put:
#.nf
#
#	$Mask{'A'} = $Mask{'B'} = $Mask{'C'};
#.fi
#
#	in .updrevrc.pl whereas more complex sub-netting requires
#	something like:
#.nf
#
#	&netMasks('/etc/netmasks');
#.fi
#
#	in .updrevrc.pl to populate %Mask with all the known masks.
#
#	When networks are sub-netted on other than 8bit boundaries,
#	things get more messy.  Consider 192.168.168.128/26, the owner
#	of 192.168.168.0 would have the following in the zone file for
#	168.168.in-addr.arpa (rev/192.168.168.rev):
#.nf
#
#	128	IN	NS	ns.some.site.
#		IN	NS	ns2.some.site.
#	129	IN	CNAME	129.128.168.168.192.in-addr.arpa.
#	130	IN	CNAME	130.128.168.168.192.in-addr.arpa.
#	::
#	::
#.fi
#
#	Note that because of the mask, 'updrev.pl' will not touch that
#	file and so the above can go in the '.rev' file rather than a
#	'.inc' file and 'gencidrzone' by Mathias Koerber can be used to
#	generate all the CNAME records.
#
#	The zone file for 128.168.168.192.in-addr.arpa
#	(rev/192.168.168.128.rev) which would be delegated to
#	"some.site" would contain the normal PTR records:
#.nf
#
#	129	IN	PTR	host129.some.site.
#	130	IN	PTR	host130.some.site.
#	::
#	::
#.fi
#
#	From the above, it is perhaps clear why we ignore sub-netting
#	where possible.  The default behaviour is rarely a problem
#	since while many sites sub-net their network, other than ISPs,
#	few delegate portions of their in-addr.arpa zones.
#
# SEE ALSO:
#	getdata.pl(1), install_new.sh(1).
#
# BUGS:
#	Much of the simplification arises from supporting only the DNS
#	setup described in http://www.crufty.net/help/dns/dnsmagic.html
#	
# AUTHOR:
#	Simon J. Gerraty <sjg@crufty.net>
#

# RCSid:
#	$Id: updrev.pl,v 1.33 2004/12/28 23:35:44 sjg Exp $
#
#	@(#)Copyright (c) 1997-2001 Simon J. Gerraty.
#	All rights reserved.
#	
#	This software is the property of Simon J. Gerraty.
#	It may be used only in accordance with the conditions outlined
#	in the accompanying LICENSE file. If your distribution does
#	not contain a LICENSE file, then you have no rights to use
#	this software and should contact sales@crufty.net to obtain a
#	licensed distribution.
#	
#	This software is provided "AS IS" in the hope that it will be
#	of use.  Since it is not possible to anticipate all possible
#	use of this software, there is absolutely NO WARRANTY.
#
#	Please forward bug reports and fixes to help@crufty.net
#

if ($0 =~ m,^(.*)/([^/]+)$,) {
  $Mydir = $1;
  $Myname = $2;
} else {
  $Mydir = '.';
  $Myname = $0;
}

push(@INC, $Mydir);
$ENV{'PATH'} = "$Mydir:$ENV{'PATH'}";
$opt_n=$ENV{'HOSTNAME'};
$opt_R='rev';
$opt_N='named.boot';
$opt_c='/dev/null';
$opt_A=$ENV{'TOP_ARPA_ZONE'};
$opt_T=$ENV{'TOP_ZONE'};

$PTR = "/tmp/.ptr$$";
$OPTR = "/tmp/.optr$$";

# just to show what should be in here...
%Mask = ('A', '0xff000000',
	 'B', '0xffff0000',
	 'C', '0xffffff00',
	 );

$SkipRC = 1;
require "$Mydir/funcs.pl";
require 'cidr_subs.pl';

&source('.dnsrc.pl');
&source(1,'.updrevrc.pl','etc',$Mydir,'.');

require 'getopts.pl';

&Getopts('d:D:A:T:R:N:c:n:a:z:');

do $opt_c if (-s $opt_c);

$opt_z = 'db' if ($opt_z eq '');

sub lastWord {
  local(@w) = split(/\s+/, $_[0]);
  $w[$#w];
}

$opt_N = &lastWord($opt_N);
$opt_R = &lastWord($opt_R);

# might be called by .updrevrc.pl
sub netMasks {
  local($file) = @_;

  $file = '/etc/netmasks' if ($file eq '');
  if (open(M, "< $file")) {
    while (<M>) {
      next unless (m/^\d/o);
      chop;
      ($n,$m,$junk) = split(/\s+/);
      $Mask{$n} = $m;
    }
    close M;
  }
}

sub getTail {
  local($d) = @_;

  if ($d =~ m/^[^.]+\.(.+)/) {
    $d = $1;
  } else {
    $d = '.';
  }
  $d;
}

sub check_parent_soa {
  local($op,$T,$t) = @_;
  local($s);
  local($p) = $op;
  local($rT) = $T;
  
  $rT = $t if ($T eq '');

  $rT =~ s/\.+$//;
  $rT =~ s/\./\\./g;
  $rT =~ s/\*/.*/g;

  for ($s = ''; $s eq '' && $p =~ m/$rT$/i; $p = &getTail($p)) {
    $s = $SOA{$p};
  }
  &dbgErr($DBG{'soa'}, 5, "check_parent_soa($op,$T,$t): $p = $s\n");
  $s;
}

# we come here with [4.]3.2.1.in-addr.arpa
sub check_inaddr_soa {
  local($n2) = @_;
  local(@a) = split(/\./, $n2);
  local($n,$s,$x);

  if ($a[$#a] eq 'arpa') {
    $x = $#a - 2;
    if ($a[$x] < 128) {
      $n = join('.', @a[$x .. $#a]);
    } elsif ($a[2] < 192) {
      $n = join('.', @a[$x - 1 .. $#a]);
    } else {
      $n = join('.', @a[$x - 2 .. $#a]);
    }
    # while we may be authoritative for $n,
    # we always first check for delegated $n2.
    $s = &check_parent_soa($n2, $opt_A, $n);
  } else {
    $s = 0;			# ignore it please
  }
  &dbgErr($DBG{'soa'}, 7, "check_inaddr_soa($n2) = $s\n");
  $s;
}

sub setup {
  $opt_n =~ s/\.,/,/g;
  $opt_n =~ s/\.$//;
  $opt_n =~ y/A-Z/a-z/;

  if ($opt_d ne '') {
    foreach $x (split(/,/, $opt_d)) {
      ($k,$v) = split(/[\.:=]/, $x);
      $DBG{$k} = $v | 1;
    }
  }
  if ($DBG{'sh'} ne '') {
    $set_x = 'set -x;';
    $sh_x = 'sh -x';
  } else {
    $set_x = '';
    $sh_x = '';
  }

  foreach $n (split(/,/, $opt_n)) {
    $NS{$n} = 1;
    &dbgErr($DBG{'setup'}, 0, "setup: NS $n\n");
    if ($opt_T eq '') {
      $opt_T = &getTail($n);
    }
  }
  &dbgErr($DBG{'setup'}, 0, "setup: opt_T=$opt_T\n");
  
  system("$set_x mkdir -p $opt_R && find $opt_R -name '*.$opt_z.new' -print | xargs rm -f");

  if ($opt_D ne '' && -s $opt_D) {
    open(STDIN, "< $opt_D") || die;
    &dbgErr($DBG{'setup'}, 0, "setup: data=$opt_D\n");
  } else {
    open(STDIN, "$Mydir/getdata.pl @ARGV |") || die;
    &dbgErr($DBG{'setup'}, 0, "setup: data=$Mydir/getdata.pl @ARGV |\n");
  }
  open(PTR, "> $PTR") || die;
}

# this is an example of what can be done to skip existing PTR
# records for some domains or addresses, skip_fqdn is perhaps more
# useful.
# .updrevrc.pl can set $PTR_hook='skip_ptr' and
# populate @Skip_ptr_re or %Skip_ptr_a or both.

sub skip_ptr {
  local($a,$fqdn) = @_;
  local($s);
  
  if (($s = $Skip_ptr_a{$a}) eq '') {
    foreach $e (@Skip_ptr_re) {
      if ($fqdn =~ m/$e/) {
	&dbgErr($DBG{'skip_ptr'}, 0, "skipping $fqdn -> /$e/i\n");
	$s = $Skip_ptr_a{$a} = $fqdn;
	last;
      }
    }
  }
  ($s eq '');
}

# sometimes we want to skip a domain or certain names.
sub skip_fqdn {
  local($a,$fqdn) = @_;
  local($s);
  
  if (($s = $Skip_fqdn{$fqdn}) eq '') {
    foreach $e (@Skip_fqdn_re) {
      if ($fqdn =~ m/$e/i) {
	&dbgErr($DBG{'skip_fqdn'}, 0, "skipping $fqdn -> /$e/i\n");
	$s = $Skip_fqdn{$fqdn} = $fqdn;
	last;
      }
    }
  }
  ($s eq '');
}

sub revA {
  local(@a) = @_;
  local($i, $j, @r);

  for ($i = 0, $j = $#a; $i < scalar(@a); --$j, ++$i) {
    $r[$i] = $a[$j];
  }
  @r;
}

sub get_mask {
  local($a) = @_;
  local($mask) = '';
  local($x,$dmask,@k);
  local(@a) = split(/\./, $a);
  
  if ($a[0] < 128) {
    @k = ($a[0], "$a[0].0.0.0");
    $dmask = $Mask{'A'} || '0xff000000';
  } elsif ($a[0] < 192) {
    @k = ("$a[0].$a[1]", "$a[0].$a[1].0.0");
    $dmask = $Mask{'B'} || '0xffff0000';
  } else {
    @k = ("$a[0].$a[1].$a[2]", "$a[0].$a[1].$a[2].0");
    $dmask = $Mask{'C'} || '0xffffff00';
  }

  foreach $x (@k) {
    $mask = $Mask{$x};
    last if ($mask ne '');
  }
  $mask = $dmask if ($mask eq '');

  &dbgErr($DBG{'mask'}, 1, "get_mask($a): $mask\n");

  return $mask;
}

# ensure we are consistent with rev files.
sub gen_ptr {
  local($a,$fqdn) = @_;
  local($N,$H,@a);
  local($mask) = &get_mask($a);
  local($key,$x,$bits,%cidr);

  %cidr = &CIDR($a, $mask);
  $key = $cidr{'prefix'};
  if (($N = $Prefix{$key}) eq '') {
    $N = $key;
    $bits = $cidr{'bits'};
    # how many octets do we want for $H?
    $x = int((32 - $bits) / 8);
    # if its an 8bit boundary, we lose the trailing .0's
    # otherwise we retain them - to differentiate b/w the parent
    # zone and the first sub-net.
    if (($bits % 8) == 0) {
      $N =~ s/(\.0)+$//;
      $x -= 1;
    }
    $HostBytes{$key} = $x;
    # now simply reverse it and remember it.
    $Prefix{$key} = $N = join('.', &revA(split(/\./, $N))) . ".in-addr.arpa";
  } else {
    $x = $HostBytes{$key};
  }
  @a = &revA(split(/\./, $a));
  if ($x < 1) {
    $H = $a[0];
  } else {
    $H = join('.', @a[0 .. $x]);
  }
  &dbgErr($DBG{'ptr'}, 4, "gen_ptr($a,$fqdn): $N $H\n");

  return ($N, $H, "$fqdn.");
}


sub getptrs {
  local(*I) = @_;

  &dbgErr($DBG{'gen'}, 0, "getptrs:\n");

  while (<I>) {
    chop;
    next if (m/localhost/);
    next if (m/\s1.0.0.127.in-addr/); # some folk call it something else!
    
    s/\./ /;
    ($type,$key,$origin,$data,$xtra) = split(/\s+/, $_, 5);

    if ($type eq 'SOA') {
      # the expression by itself does not work with perl5!
      $SOA{"$key.$origin"} = ($NS{$data} ne '') ? 1 : 0;
      next;
    }
    if ($type eq 'PTR') {
      $s = &check_inaddr_soa($origin);
      next if ($s != 1);
      $a = "$key.$origin";
      $a =~ s/\.in-addr.*//;
      @a = split(/\./, $a);
      if (($n = scalar(@a)) > 4) {
	# must be a CIDR PTR we want the 1st N - 4 
	# and the last N - (4 - (N - 4)) octets.
	$x = $n - 5;
	$y = $n - (3 - $x);
	@a = ($a[0 .. $x], @a[$y .. $#a]);
      }
      $a = join('.', &revA(@a));
      if ($PTR_hook ne '') {
	next unless &$PTR_hook($a, $data);
      }
      @p = &gen_ptr($a, $data);
      $OldP{$a} = "@p";
      next;
    }
    if ($type eq 'A') {
      next if ($data eq '127.0.0.1'); # !
      $p = $OldP{$data};
      if ($p ne '') {
	if ($p =~ m/\s$key\.$origin\.$/) {
	  # ok old PTR is valid, keep it.
	  print PTR "$p\n";
	  $P{$data} = '';	# flush any other candidate
	  $A{$data} = ';';	# just in case we were first
	  next;
	}
	# ok we don't know if PTR is valid (yet).
      }
      if ($A_hook ne '') {
	next unless &$A_hook($data, "$key.$origin");
      }
      $c = $A{$data};
      $A{$data} = ';';		# for next time

      #
      # We may be authoritative for "origin" but
      # not for the IN-ADDR zone, so check for that
      # first.
      #
      @a = split(/\./, $data);
      $N = join('.', &revA(@a)) . ".in-addr.arpa";
      $s = &check_inaddr_soa($N);
      if ($s eq '') {
	# Ok, no one authoritative for in-addr zone,
	# so check origin.
	$s = &check_parent_soa($origin, $opt_T);
	if ($s eq '') {
	  # no one owns it!?
	  next if ($UNCLAIMED_REVZONE ne 'take');
	}
      }
      next if ($s eq '0');
      if ($A_hook ne '') {
	next unless &$A_hook($data, "$key.$origin");
      }
      @p = &gen_ptr($data, "$key.$origin");
      if ($c eq '' && $p ne '') {
	# ok, we have a candidate PTR, but
	# it is not the same as an old PTR and
	# we don't (yet) know which is correct
	# so just remember it for now.
	$P{$data} = "@p";
	next;
      }
      print PTR "$p[0] $c$p[1] $p[2]\n";
      next;
    }
  }
  # time to deal with candidate PTRs that are not
  # the same as old ones.  Any candidate PTR that
  # is still there needs to be printed.
  foreach $data (keys(%P)) {
    if (($p = $P{$data}) ne '') {
      print PTR "$p\n";
    }
  }
}

sub genrevs {
  $last='';
  $RZONE='';

  &dbgErr($DBG{'gen'}, 0, "genrevs:\n");

  $nonPTRre="[ \t]PTR[ \t]";
  
  if (-s $opt_N) {
    system("$set_x cp $opt_N $opt_N.new");
  }
  if ($opt_a ne '') {
    open(ADDED, "> $opt_a") || die;
  }
  # the sort -u suppresses duplicates.
  # the sort -n +1 -2 means the final file is in numeric order
  # though commented out entries end up at the top
  # rather than the bottom.
  if ($updrev_no_sort_numeric eq '') {
    open(STDIN, "sort -u $PTR | sort -n +1 -2 |") || die;
  } else {
    # this is the old behavior.
    open(STDIN, "sort -u $PTR |") || die;
  }
  open(OUT, ">> $opt_N.new") || die;

  %Opened = ();
  
  while (<STDIN>) {
    ($origin, $n, $fqdn) = split(/\s+/);

    if ($last ne $origin) {
      @a = split(/\./, $origin);
      $RZONE='';

      foreach $x (@a) {
	last if ($x eq 'in-addr');
	$RZONE="$x.$RZONE";
      }
      $RZONE="$opt_R/$RZONE";
      $file = "${RZONE}soa";
      chop($old = `grep -i "[^0-9]$origin" "$opt_N.new"`);
      if ($old eq '') {
	if ($opt_N =~ m/name.*boot/) {
	  # bind-4
	  print OUT "primary $origin $file\n";
	} else {
	  # bind-8
	  print OUT "
zone \"$origin\" {
	type master;
	file \"$file\";
};\n\n";
	}
	if ($opt_a ne '') {
	  print ADDED "${RZONE}$opt_z\n";
	}
      }
      if ($Opened{$origin} eq '') {
	$Opened{$origin} = 1;
	&dbgErr($DBG{'gen'}, 1, "opening ${RZONE}$opt_z.new\n");
	open(RZONE, "> ${RZONE}$opt_z.new") || die;
	print RZONE "; WARNING: This file is automagically maintained.
; Place all non-PTR records in ${RZONE}inc and \$INCLUDE it
; or they will dissappear.
;\n";
	if (-s "${RZONE}inc") {
	  print RZONE "\$INCLUDE ${RZONE}inc\n;\n";
	}
      } else {
	&dbgErr($DBG{'gen'}, 1, "appending to ${RZONE}$opt_z.new\n");
	open(RZONE, ">> ${RZONE}$opt_z.new") || die;
      }
    }
    $last=$origin;
    &dbgErr($DBG{'gen'}, 1, "adding '$n\tIN\tPTR\t$fqdn' to $origin\n");
    
    print RZONE "$n\tIN\tPTR\t$fqdn\n";
  }
  if ($opt_a ne '') {
    close ADDED;
  }
  close RZONE;
  close OUT;
}

sub cleanup {
  &dbgErr($DBG{'gen'}, 0, "cleanup:\n");
  system("$set_x find $opt_N.new $opt_R -name '*.new' -print 2> /dev/null | $sh_x $Mydir/install_xnew.sh");
  unlink($PTR);
}

sub main {
  &setup;
  &getptrs(STDIN);
  close PTR;
  &genrevs;
  &cleanup;
}

&main;
exit 0;

